#!/usr/bin/env python3
# Copyright 2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import requests
from functools import partial
from typing import Optional, Callable, List

import rich
import click

VOLARE_REPO_OWNER = os.getenv("VOLARE_REPO_OWNER") or "efabless"
VOLARE_REPO_NAME = os.getenv("VOLARE_REPO_NAME") or "volare"
VOLARE_REPO_ID = f"{VOLARE_REPO_OWNER}/{VOLARE_REPO_NAME}"
VOLARE_REPO_HTTPS = f"https://github.com/{VOLARE_REPO_ID}"
VOLARE_REPO_API = f"https://api.github.com/repos/{VOLARE_REPO_ID}"

SKY130_VARIANTS = ["sky130A", "sky130B"]
SKY130_DEFAULT_LIBRARIES = [
    "sky130_fd_sc_hd",
    "sky130_fd_sc_hvl",
    "sky130_fd_io",
    "sky130_fd_pr",
]

opt = partial(click.option, show_default=True)


def opt_pdk_root(function: Callable):
    function = click.option(
        "--pdk-root",
        required=False,
        default=os.getenv("PDK_ROOT") or os.path.expanduser("~/.volare"),
        help="Path to the PDK root",
        show_default=True,
    )(function)
    return function


def opt_build(function: Callable):
    function = opt(
        "-l",
        "--include-libraries",
        multiple=True,
        default=SKY130_DEFAULT_LIBRARIES,
        help="Libraries to include in the build. You can use -l multiple times to include multiple libraries. Pass 'all' to include all of them.",
    )(function)
    function = opt(
        "-j",
        "--jobs",
        default=1,
        help="Specifies the number of commands to run simultaneously.",
    )(function)
    function = opt("--sram/--no-sram", default=True, help="Enable or disable sram")(
        function
    )
    function = opt(
        "--clear-build-artifacts/--keep-build-artifacts",
        default=False,
        help="Whether or not to remove the build artifacts. Keeping the build artifacts is useful when testing.",
    )(function)
    return function


def opt_push(function: Callable):
    function = opt("-o", "--owner", default=VOLARE_REPO_OWNER, help="Repository Owner")(
        function
    )
    function = opt("-r", "--repository", default=VOLARE_REPO_NAME, help="Repository")(
        function
    )
    function = opt(
        "-t",
        "--token",
        required=(os.getenv("GITHUB_TOKEN") is None),
        default=os.getenv("GITHUB_TOKEN"),
        help="Github Token",
    )(function)
    return function


def check_version(
    version: Optional[str],
    tool_metadata_file_path: Optional[str],
    console: Optional[rich.console.Console] = None,
) -> str:
    """
    Takes an optional version and tool_metadata_file_path.

    If version is set, it is returned.

    If not, tool_metadata_file_path is checked if it exists.

    If not specified, ./tool_metadata.yml and ./dependencies/tool_metadata.yml
    are both checked if they exist.

    If none are specified, execution is halted.

    Otherwise, the resulting metadata file is parsed for an open_pdks version,
    which is then returned.
    """
    if version is not None:
        return version

    def pr(*args):
        if console is not None:
            console.log(*args)
        else:
            print(*args)

    import yaml

    if tool_metadata_file_path is None:
        tool_metadata_file_path = os.path.join(".", "tool_metadata.yml")
        if not os.path.isfile(tool_metadata_file_path):
            tool_metadata_file_path = os.path.join(
                ".", "dependencies", "tool_metadata.yml"
            )
            if not os.path.isfile(tool_metadata_file_path):
                pr(
                    "Any of ./tool_metadata.yml or ./dependencies/tool_metadata.yml not found. You'll need to specify the file path or the commits explicitly."
                )
                exit(os.EX_USAGE)

    tool_metadata = yaml.safe_load(open(tool_metadata_file_path).read())

    open_pdks_list = [tool for tool in tool_metadata if tool["name"] == "open_pdks"]

    if len(open_pdks_list) < 1:
        pr("No entry for open_pdks found in tool_metadata.yml")
        exit(os.EX_USAGE)

    version = open_pdks_list[0]["commit"]

    pr(f"Found version {version} in {tool_metadata_file_path}.")

    return version


def get_volare_dir(pdk_root: str) -> str:
    return os.path.join(pdk_root, "volare", "sky130")


def get_versions_dir(pdk_root: str) -> str:
    return os.path.join(get_volare_dir(pdk_root), "versions")


def get_version_dir(pdk_root: str, version: str) -> str:
    return os.path.join(get_versions_dir(pdk_root), version)


def get_link_of(version: str) -> str:
    return f"{VOLARE_REPO_HTTPS}/releases/download/sky130-{version}/default.tar.xz"


def get_version_list() -> List[str]:
    response_str = requests.get(f"{VOLARE_REPO_API}/releases").content.decode("utf8")
    releases = json.loads(response_str)
    pdk_versions = [release["tag_name"] for release in releases]
    pdk_versions_by_pdk = {}
    for version in pdk_versions:
        pdk, hash = version.split("-")
        if pdk_versions_by_pdk.get(pdk) is None:
            pdk_versions_by_pdk[pdk] = pdk_versions_by_pdk.get(pdk) or []
        pdk_versions_by_pdk[pdk].append(hash)
    return pdk_versions_by_pdk["sky130"]



#!/usr/bin/env python3
# Copyright 2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import json
import uuid
import tarfile
import pathlib
import requests
import tempfile

import rich
import rich.tree
import click
import rich.progress
from click_default_group import DefaultGroup

from .git_multi_clone import mkdirp
from .common import (
    get_link_of,
    opt,
    opt_build,
    opt_push,
    opt_pdk_root,
    check_version,
    get_versions_dir,
    get_version_dir,
    get_volare_dir,
    SKY130_VARIANTS,
    get_version_list,
)
from .build import build, push


def get_installed_list(pdk_root):
    versions_dir = get_versions_dir(pdk_root)
    mkdirp(versions_dir)
    return os.listdir(versions_dir)


def print_installed_list(pdk_root, console):
    installed_list = get_installed_list(pdk_root)
    versions_dir = get_versions_dir(pdk_root)
    if len(installed_list) == 0:
        console.print("[red]No PDKs installed.")
        return

    version = get_current_version(pdk_root)

    tree = rich.tree.Tree(f"{versions_dir}")
    for installed in installed_list:
        if installed == version:
            tree.add(f"[green][bold]{installed} (enabled)")
        else:
            tree.add(installed)
    console.print(tree)


def print_remote_list(pdk_root, console, pdk_list):
    installed_list = get_installed_list(pdk_root)

    version = get_current_version(pdk_root)

    tree = rich.tree.Tree("Pre-built PDKs")
    for pdk in pdk_list:
        if pdk == version:
            tree.add(f"[green][bold]{pdk} (enabled)")
        elif pdk in installed_list:
            tree.add(f"[green]{pdk} (installed)")
        else:
            tree.add(pdk)
    console.print(tree)


def get_current_version(pdk_root):
    current_file = os.path.join(get_volare_dir(pdk_root), "current")
    current_file_dir = os.path.dirname(current_file)
    mkdirp(current_file_dir)
    pathlib.Path(current_file).touch(exist_ok=True)

    return open(current_file).read().strip()


@click.group(cls=DefaultGroup, default="output", default_if_no_args=True)
def manage():
    """Allows you to list and enable installed PDKs. (default)"""
    pass


@click.command()
@opt_pdk_root
def output_cmd(pdk_root):
    """(Default) Outputs the currently installed PDK version."""

    if sys.stdout.isatty():
        console = rich.console.Console()
        print_installed_list(pdk_root, console)
    else:
        version = get_current_version()
        if version == "":
            exit(1)
        else:
            print(version, end="")


@click.command("list", hidden=True)
@opt_pdk_root
def list_cmd(pdk_root):
    """Lists PDK versions that are remotely available. JSON if not outputting to a tty."""

    pdk_versions = get_version_list()

    if sys.stdout.isatty():
        console = rich.console.Console()
        print_remote_list(pdk_root, console, pdk_versions)
    else:
        print(json.dumps(pdk_versions))


@click.command("path")
@opt_pdk_root
@click.argument("version", required=False)
def path_cmd(pdk_root, version):
    """Prints the path of a specific pdk version installation."""
    path_to_print = pdk_root
    if version is not None:
        path_to_print = os.path.join(get_versions_dir(pdk_root), version)
    print(path_to_print, end="")


def enable(
    pdk_root,
    version,
    build_if_not_found=False,
    also_push=False,
    build_kwargs={},
    push_kwargs={},
):
    console = rich.console.Console()

    current_file = os.path.join(get_volare_dir(pdk_root), "current")
    current_file_dir = os.path.dirname(current_file)
    mkdirp(current_file_dir)

    version_directory = get_version_dir(pdk_root, version)

    version_paths = [
        os.path.join(version_directory, variant) for variant in SKY130_VARIANTS
    ]
    final_paths = [os.path.join(pdk_root, variant) for variant in SKY130_VARIANTS]

    if not os.path.exists(version_directory):
        link = get_link_of(version)
        status = requests.head(link).status_code
        if status == 404:
            if build_if_not_found:
                console.print(
                    f"Version {version} not found either locally or remotely. Will attempt to build."
                )
                build(pdk_root=pdk_root, version=version, **build_kwargs)
                if also_push:
                    push(pdk_root, version, **push_kwargs)
            else:
                console.print(
                    f"[red]Version {version} not found either locally or remotely.\nTry volare build {version}."
                )
                exit(1)
        else:
            tempdir = tempfile.gettempdir()
            tarball_directory = os.path.join(tempdir, "volare", f"{uuid.uuid4()}")
            mkdirp(tarball_directory)

            tarball_path = os.path.join(tarball_directory, f"{version}.tar.xz")
            with requests.get(link, stream=True) as r:
                with rich.progress.Progress() as p:
                    task = p.add_task(
                        f"Downloading pre-built tarball for {version}…",
                        total=int(r.headers["Content-length"]),
                    )
                    r.raise_for_status()
                    with open(tarball_path, "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            p.advance(task, advance=len(chunk))
                            f.write(chunk)

                    task = p.add_task("Unpacking…")
                    with tarfile.open(tarball_path, mode="r:*") as tf:
                        p.update(task, total=len(tf.getmembers()))
                        for i, file in enumerate(tf.getmembers()):
                            p.update(task, completed=i + 1)
                            final_path = os.path.join(version_directory, file.name)
                            final_dir = os.path.dirname(final_path)
                            mkdirp(final_dir)
                            with tf.extractfile(file) as io:
                                with open(final_path, "wb") as f:
                                    f.write(io.read())

                    for variant in SKY130_VARIANTS:
                        variant_install_path = os.path.join(version_directory, variant)
                        variant_sources_file = os.path.join(
                            variant_install_path, "SOURCES"
                        )
                        if not os.path.isfile(variant_sources_file):
                            with open(variant_sources_file, "w") as f:
                                print(f"open_pdks {version}", file=f)

                    os.unlink(tarball_path)

    with console.status(f"Enabling version {version}…"):
        for path in final_paths:
            if os.path.exists(path):
                if os.path.islink(path):
                    os.unlink(path)
                else:
                    console.print(
                        f"[red]Error: {path} exists, and not as a symlink. Please manually remove it before continuing."
                    )
                    exit(1)

        for vpath, fpath in zip(version_paths, final_paths):
            src = os.path.relpath(vpath, pdk_root)
            os.symlink(src=src, dst=fpath)

        with open(current_file, "w") as f:
            f.write(version)

    console.print(f"PDK version {version} enabled.")


@click.command("enable")
@opt_pdk_root
@click.option(
    "-f",
    "--metadata-file",
    "tool_metadata_file_path",
    default=None,
    help="Explicitly define a tool metadata file instead of searching for a metadata file",
)
@click.argument("version", required=False)
def enable_cmd(pdk_root, tool_metadata_file_path, version):
    """
    Activates a given PDK version.

    Parameters: <version> (Optional)

    If a version is not given, and you run this in the top level directory of
    tools with a tool_metadata.yml file, for example OpenLane or DFFRAM,
    the appropriate version will be enabled automatically.
    """
    console = rich.console.Console()
    version = check_version(version, tool_metadata_file_path, console)
    enable(pdk_root, version)


@click.command("enable_or_build", hidden=True)
@opt_pdk_root
@opt_push
@opt_build
@opt("--also-push/--dont-push", default=False, help="Also push.")
@click.option(
    "-f",
    "--metadata-file",
    "tool_metadata_file_path",
    default=None,
    help="Explicitly define a tool metadata file instead of searching for a metadata file",
)
@click.argument("version")
def enable_or_build_cmd(
    include_libraries,
    jobs,
    sram,
    pdk_root,
    owner,
    repository,
    token,
    clear_build_artifacts,
    tool_metadata_file_path,
    also_push,
    version,
):
    """
    Attempts to activate a given PDK version. If the version is not found locally or remotely,
    it will instead attempt to build said version.

    Parameters: <version>
    """
    console = rich.console.Console()
    version = check_version(version, tool_metadata_file_path, console)
    enable(
        pdk_root,
        version,
        build_if_not_found=True,
        also_push=also_push,
        build_kwargs={
            "include_libraries": include_libraries,
            "jobs": jobs,
            "sram": sram,
            "clear_build_artifacts": clear_build_artifacts,
        },
        push_kwargs={"owner": owner, "repository": repository, "token": token},
    )



#!/usr/bin/env python3
# Copyright 2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import venv
import uuid
import shutil
import pathlib
import tarfile
import tempfile
import subprocess
from concurrent.futures import ThreadPoolExecutor

import rich
import click
from rich.progress import Progress

from .git_multi_clone import GitMultiClone, Repository, mkdirp
from .common import (
    opt_pdk_root,
    opt_build,
    check_version,
    get_version_dir,
    get_volare_dir,
    SKY130_VARIANTS,
    SKY130_DEFAULT_LIBRARIES,
    VOLARE_REPO_OWNER,
    VOLARE_REPO_NAME,
    opt_push,
)


class RepoMetadata(object):
    def __init__(self, repo, default_commit, default_branch="main"):
        self.repo = repo
        self.default_commit = default_commit
        self.default_branch = default_branch


RepoMetadata.by_name = {
    "open_pdks": RepoMetadata(
        "https://github.com/RTimothyEdwards/open_pdks",
        "34eeb2743e99d44a21c2cedd467675a2e0f3bb91",
        "master",
    ),
    "sky130": RepoMetadata(
        "https://github.com/google/skywater-pdk",
        "f70d8ca46961ff92719d8870a18a076370b85f6c",
        "main",
    ),
    "magic": RepoMetadata(
        "https://github.com/RTimothyEdwards/magic",
        "d98645afc1498a41db7aff4b2e100f27e0b0bd9b",
        "master",
    ),
}


def get_open_pdks(version, build_directory, jobs=1):
    try:
        console = rich.console.Console()

        open_pdks_repo = None

        with Progress() as progress:
            with ThreadPoolExecutor(max_workers=jobs) as executor:
                gmc = GitMultiClone(build_directory, progress)
                open_pdks = RepoMetadata.by_name["open_pdks"]
                open_pdks_future = executor.submit(
                    GitMultiClone.clone,
                    gmc,
                    open_pdks.repo,
                    version,
                    open_pdks.default_branch,
                )
                open_pdks_repo = open_pdks_future.result()

        console.log(f"Done fetching {open_pdks_repo.name}.")

    except subprocess.CalledProcessError as e:
        print(e)
        print(e.stderr)
        exit(os.EX_DATAERR)


def get_sky130(include_libraries, build_directory, jobs=1):
    try:
        all = "all" in include_libraries
        console = rich.console.Console()

        sky130_repo = None
        sky130_submodules = []

        with Progress() as progress:
            with ThreadPoolExecutor(max_workers=jobs) as executor:
                gmc = GitMultiClone(build_directory, progress)
                sky130 = RepoMetadata.by_name["sky130"]
                # TODO: Get sky130 commit from open_pdks
                sky130_fut = executor.submit(
                    GitMultiClone.clone,
                    gmc,
                    sky130.repo,
                    sky130.default_commit,
                    sky130.default_branch,
                )
                sky130_repo = sky130_fut.result()
                sky130_submodules = (
                    subprocess.check_output(
                        ["find", "libraries", "-type", "d", "-name", "latest"],
                        stderr=subprocess.PIPE,
                        cwd=sky130_repo.path,
                    )
                    .decode("utf8")
                    .strip()
                    .split("\n")
                )
                sky130_submodules_filtered = [
                    sm
                    for sm in sky130_submodules
                    if (sm.split("/")[1] in include_libraries or all)
                ]
                for submodule in sky130_submodules_filtered:
                    executor.submit(
                        GitMultiClone.clone_submodule, gmc, sky130_repo, submodule
                    )
        console.log("Done fetching repositories.")

    except subprocess.CalledProcessError as e:
        print(e)
        print(e.stderr)
        exit(os.EX_DATAERR)


def build_sky130_timing(build_directory, jobs=1):
    try:
        console = rich.console.Console()
        sky130_repo = Repository.from_path(
            os.path.join(build_directory, "skywater-pdk")
        )
        sky130_submodules = (
            subprocess.check_output(
                ["find", "./libraries", "-type", "d", "-name", "latest"],
                stderr=subprocess.PIPE,
                cwd=sky130_repo.path,
            )
            .decode("utf8")
            .strip()
            .split("\n")
        )

        venv_path = os.path.join(build_directory, "venv")

        with console.status("Building venv…"):
            venv_builder = venv.EnvBuilder(with_pip=True)
            venv_builder.create(venv_path)
        console.log("Done building venv.")

        mkdirp("/tmp/volare/logs")

        with console.status("Installing python-skywater-pdk in venv…"), open(
            "/tmp/volare/logs/venv.stdout", "w"
        ) as so, open("/tmp/volare/logs/venv.stderr", "w") as se:
            subprocess.check_call(
                [
                    "bash",
                    "-c",
                    f"""
                        set -e
                        source {venv_path}/bin/activate
                        python3 -m pip install wheel
                        python3 -m pip install {os.path.join(sky130_repo.path, 'scripts', 'python-skywater-pdk')}
                    """,
                ],
                stdout=so,
                stderr=se,
            )
        console.log("Done setting up venv.")

        def do_submodule(submodule: str):
            submodule_cleaned = submodule.strip("/.").replace("/", "_")
            console.log(f"Processing {submodule}…")
            with open(
                f"/tmp/volare/logs/{submodule_cleaned}.timing.stdout", "w"
            ) as so, open(
                f"/tmp/volare/logs/{submodule_cleaned}.timing.stderr", "w"
            ) as se:
                subprocess.check_call(
                    [
                        "bash",
                        "-c",
                        f"""
                            set -e
                            source {venv_path}/bin/activate
                            cd {sky130_repo.path}
                            python3 -m skywater_pdk.liberty {submodule}
                            python3 -m skywater_pdk.liberty {submodule} all
                            python3 -m skywater_pdk.liberty {submodule} all --ccsnoise
                        """,
                    ],
                    stdout=so,
                    stderr=se,
                )
            console.log(f"Done with {submodule}.")

        with ThreadPoolExecutor(max_workers=jobs) as executor:
            for submodule in sky130_submodules:
                submodule_path = os.path.join(sky130_repo.path, submodule)
                if (
                    not os.path.exists(os.path.join(submodule_path, "cells"))
                    or "_sc" not in submodule
                ):
                    continue
                executor.submit(
                    do_submodule,
                    submodule,
                )
        console.log("Created timing data.")

    except subprocess.CalledProcessError as e:
        print(e)
        print(e.stderr)
        exit(os.EX_DATAERR)


def build_sky130(sram, build_directory, jobs=1):
    try:
        console = rich.console.Console()

        magic_tag = RepoMetadata.by_name["magic"].default_commit

        # TODO: Get magic version from open_pdks
        magic_image = f"efabless/openlane-tools:magic-{magic_tag}-centos-7"

        subprocess.check_call(["docker", "pull", magic_image])

        docker_ids = set()

        def docker_run(*args):
            nonlocal docker_ids
            container_id = str(uuid.uuid4())
            docker_ids.add(container_id)
            args = list(args)
            pdk_root_abs = os.path.abspath(build_directory)
            subprocess.check_call(
                [
                    "docker",
                    "run",
                    "--name",
                    container_id,
                    "--rm",
                    "-e",
                    f"PDK_ROOT={pdk_root_abs}",
                    "-v",
                    f"{pdk_root_abs}:{pdk_root_abs}",
                    "-w",
                    f"{pdk_root_abs}",
                    magic_image,
                ]
                + args
            )
            docker_ids.remove(container_id)

        sram_opt = "--enable-sram-sky130" if sram else ""

        interrupted = None
        try:
            console.log("Configuring open_pdks…")
            docker_run(
                "sh",
                "-c",
                f"""
                    set +e
                    cd open_pdks
                    ./configure --enable-sky130-pdk=$PDK_ROOT/skywater-pdk/libraries {sram_opt}
                """,
            )
            console.log("Done.")

            console.log("Building prequisites…")
            docker_run(
                "sh",
                "-c",
                f"""
                    set +e
                    cd open_pdks/sky130
                    make -j{jobs} prerequisites
                """,
            )
            console.log("Done.")

            console.log("Building sky130A/B…")
            docker_run(
                "sh",
                "-c",
                f"""
                    set +e
                    cd open_pdks/sky130
                    export LC_ALL=en_US.UTF-8
                    make -j{jobs}
                    make SHARED_PDKS_PATH=$PDK_ROOT install
                """,
            )
        except KeyboardInterrupt as e:
            interrupted = e
            console.log("Stopping on keyboard interrupt…")
            console.log("Killing docker containers…")
            for id in docker_ids:
                subprocess.call(["docker", "kill", id])
        console.log("Attempting to fix ownership…")
        docker_run(
            "sh",
            "-c",
            """
                set +e
                OWNERSHIP="$(stat -c "%u:%g" $PDK_ROOT)"
                chown -R $OWNERSHIP $PDK_ROOT
            """,
        )
        if interrupted is not None:
            raise interrupted
        else:
            console.log("Done.")

    except subprocess.CalledProcessError as e:
        print(e)
        print(e.stderr)
        exit(os.EX_DATAERR)


def install_sky130(build_directory, pdk_root, version):
    console = rich.console.Console()
    with console.status("Adding build to list of installed versions…"):
        version_directory = get_version_dir(pdk_root, version)
        if (
            os.path.exists(version_directory)
            and len(os.listdir(version_directory)) != 0
        ):
            backup_path = version_directory
            it = 0
            while os.path.exists(backup_path) and len(os.listdir(backup_path)) != 0:
                it += 1
                backup_path = get_version_dir(pdk_root, f"{version}.bk{it}")
            console.log(
                f"Build already found at {version_directory}, moving to {backup_path}…"
            )
            shutil.move(version_directory, backup_path)

        console.log("Copying…")
        mkdirp(version_directory)

        for variant in SKY130_VARIANTS:
            variant_build_path = os.path.join(build_directory, variant)
            variant_install_path = os.path.join(version_directory, variant)
            shutil.copytree(variant_build_path, variant_install_path)

    console.log("Done.")


# ---
def build(
    pdk_root,
    version,
    jobs=1,
    sram=True,
    clear_build_artifacts=True,
    include_libraries=None,
):
    if include_libraries is None:
        include_libraries = SKY130_DEFAULT_LIBRARIES
    build_directory = os.path.join(get_volare_dir(pdk_root), "build", version)
    get_open_pdks(version, build_directory, jobs)
    get_sky130(include_libraries, build_directory, jobs)
    build_sky130_timing(build_directory, jobs)
    build_sky130(sram, build_directory, jobs)
    install_sky130(build_directory, pdk_root, version)

    if clear_build_artifacts:
        shutil.rmtree(build_directory)


@click.command("build")
@opt_pdk_root
@opt_build
@click.option(
    "-f",
    "--metadata-file",
    "tool_metadata_file_path",
    default=None,
    help="Explicitly define a tool metadata file instead of searching for a metadata file",
)
@click.argument("version", required=False)
def build_cmd(
    include_libraries,
    jobs,
    sram,
    pdk_root,
    clear_build_artifacts,
    tool_metadata_file_path,
    version,
):
    """
    Builds the sky130 PDK using open_pdks.

    Parameters: <version> (Optional)

    If a version is not given, and you run this in the top level directory of
    tools with a tool_metadata.yml file, for example OpenLane or DFFRAM,
    the appropriate version will be enabled automatically.
    """

    version = check_version(version, tool_metadata_file_path, rich.console.Console())
    build(
        pdk_root=pdk_root,
        version=version,
        jobs=jobs,
        sram=sram,
        clear_build_artifacts=clear_build_artifacts,
        include_libraries=include_libraries,
    )


def push(
    pdk_root,
    version,
    owner=VOLARE_REPO_OWNER,
    repository=VOLARE_REPO_NAME,
    token=os.getenv("GITHUB_TOKEN"),
):
    console = rich.console.Console()

    version_directory = get_version_dir(pdk_root, version)
    if not os.path.isdir(version_directory):
        console.print("[red]Version not found.")
        exit(os.EX_NOINPUT)

    tempdir = tempfile.gettempdir()
    tarball_directory = os.path.join(tempdir, "volare", f"{uuid.uuid4()}", version)
    mkdirp(tarball_directory)

    tarball_path = os.path.join(tarball_directory, "default.tar.xz")

    with Progress() as progress:
        path_it = pathlib.Path(version_directory).glob("**/*")
        files = [str(path) for path in path_it if path.is_file()]
        task = progress.add_task("Compressing…", total=len(files))
        with tarfile.open(tarball_path, mode="w:xz") as tf:
            for i, file in enumerate(files):
                progress.update(task, completed=i + 1)
                path_in_tarball = os.path.relpath(file, version_directory)
                tf.add(file, arcname=path_in_tarball)
    console.log(f"Compressed to {tarball_path}.")

    tag = f"sky130-{version}"

    # If someone wants to rewrite this to not use ghr, please, by all means.
    console.log("Starting upload…")
    subprocess.check_call(
        [
            "ghr",
            "-owner",
            owner,
            "-repository",
            repository,
            "-token",
            token,
            "-body",
            f"Volare build of sky130 with Open_PDKs {version}",
            "-commitish",
            "releases",
            tag,
            tarball_path,
        ]
    )
    console.log("Done.")


@click.command("push", hidden=True)
@opt_pdk_root
@opt_push
@click.argument("version")
def push_cmd(owner, repository, token, pdk_root, version):
    """
    For maintainers: Package and release a build to the public.

    Requires ghr: github.com/tcnksm/ghr

    Parameters: <version> (required)
    """
    push(pdk_root, version, owner, repository, token)
